%%% USAGE:
% 1 - install matconvnet following instructions 
% here [http://www.vlfeat.org/matconvnet/install/]
% 2 - Download (request) the Vaihingen or Potsdam datasets
% here [http://www2.isprs.org/commissions/comm3/wg4/tests.html]
% 3 - Dowload compile and install this https://github.com/aosokin/cropRectanglesMex
% 3 - set up the folders from where to load datasets 
% 4 - run this file
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%
function [net, info] = MAIN_Vaihingen_finetune()

% file and data setup. MODIFY WITH YOUR FOLDERS
DCNN_VAIHINGEN_SET_UP


addpath(genpath('.'))
addpath(genpath('../../../external_tools/matconvnet/matlab/'))
addpath(genpath('../../../external_tools/matconvnet/examples/'))
addpath(genpath('../../../external_tools/cropRectanglesMex-master/'))
vl_setupnn();

opts.expDir             = ['cnn_models/VA_fullPatchLabelingNet/'];

%%%%%%%%%%%%%%%%%% OPTIONS

opts.dataDir            = { [PARS.dir.DIR 'images/']; 
                            [PARS.dir.IMTRAINALL PARS.dir.IMTEST];...
                            [PARS.dir.DSMTRAIN PARS.dir.DSMTEST]};

% indexes follow order in PARS.dir.IMTRAINALL = [PARS.DIR.IMTRAIN PARS.dir.IMVALID]
opts.trainIms = 1:11;
opts.valIms   = 12:16;

opts.legend   = PARS.in.legend;

opts.sizePatch          = [65];
opts.predGtPatch        = 65;
opts.resampleBatch      = 20;
opts.batchSize          = 128 ;
opts.NbatchTrain        = 500; 
opts.NbatchVal          = 50 ;

opts.gpus               = 1;
opts.currEpoch          = 1;

opts.continue           = false;
opts.continueFromIter   = []   ;

opts.prefetch           = false ;

opts.learningRate       = [repmat(1e-6,1,10)];
opts.weightDecay        = [repmat(0.0001,1,10)];
opts.momentum           = 0.9 ;
opts.numEpochs          = length(opts.learningRate);

opts.adddsm = true;

rgbMean   = [120.5403 81.7266 94.4010 80.5426]./255; %
opts.meanImage    = single(cat(3,repmat(rgbMean(1),opts.sizePatch,opts.sizePatch), ...
    repmat(rgbMean(2),opts.sizePatch,opts.sizePatch),...
    repmat(rgbMean(3),opts.sizePatch,opts.sizePatch),...
    repmat(rgbMean(4),opts.sizePatch,opts.sizePatch))); clear rgbMean

% Data sampling n(see below)
opts.sampleImdbFun = @getVaihingenImdb;
[imdb, opts] = opts.sampleImdbFun(opts);

%%%%%%%%%%%%%%%%%% NET DEFINITION: FINE TUNE
load('Vaihingen-denseprediction-distribution.mat')
net = dagnn.DagNN.loadobj(net) ;

% set proper layers (loss, errors, etc.)
net.removeLayer('softmax')

net.addLayer('objective', ...
    dagnn.SegmentationLoss('loss', 'softmaxlog'), ...
    {'prediction','label'}, 'objective') ;

net.addLayer('accuracy', ...
    dagnn.SegmentationAccuracy(), ...
    {'prediction','label'}, 'accuracy') ;

net.meta.normalization.averageImage = opts.meanImage ;
% 
% net.print({'input',[opts.sizePatch,opts.sizePatch,size(opts.meanImage,3),opts.batchSize],...
%     'label',[opts.sizePatch,opts.sizePatch,1,opts.batchSize]},...
%     'parameters','*','MaxNumColumns',10);


% feed training function
opts.train    = find(imdb.images.set == 1);
opts.val      = find(imdb.images.set == 3);

fprintf('%s\n',opts.expDir)

opts.sampleImdbFun = @getVaihingenImdb;

[net, info] = DCNN_my_dagcnn_train(net, imdb, @getBatch, opts);
end

%---------------------------------------------------------------------
function [inputs] = getBatch(imdb, batch)
% --------------------------------------------------------------------
im = imdb.images.data(:,:,:,batch) - repmat(imdb.images.data_mean,1,1,1,length(batch)) ;
labels = imdb.images.labels(:,:,:,batch) ;

coin = floor(rand(1)*3);
if coin <= 1
    im = (fliplr(im));
    labels = fliplr(labels);
elseif coin <= 2 && coin > 1
    im = (flipud(im));
    labels = flipud(labels);
end

% vl_nnspnorm(im(:,:,1:3,1))
fraction = rand(1)./100;
noise = randn(size(im));
noise = fraction.*(noise+abs(min(noise(:))));
im(:,:,1:3,:) = (im(:,:,1:3,:) + noise(:,:,1:3,:))./(max(im(:) + noise(:)));

if ~isempty(imdb.gpus)
    im = gpuArray(im) ;
end

inputs = {'input', im, 'label', labels} ;
end

% --------------------------------------------------------------------
function [imdb, opts] = getVaihingenImdb(opts)
% --------------------------------------------------------------------
% Preapre the imdb structure, returns image data with mean image subtracted

w_im = opts.sizePatch;

fprintf('%i --> ',w_im)

w_pa = opts.sizePatch;
w_gt = opts.predGtPatch;
cep  = opts.currEpoch;

n = ceil(opts.batchSize*opts.NbatchTrain/length(opts.trainIms));

fprintf('t:')
x1 = [];
y1 = [];
for imToSample = opts.trainIms
    fprintf('.')
   
    gt = single(rgb2label(imread([opts.dataDir{1} 'gts/' ...
        opts.dataDir{2}{imToSample} '.tif']),opts.legend));
    
    im = single(imread([opts.dataDir{1} 'tiles/' ...
        opts.dataDir{2}{imToSample} '.tif']))./255;
    
    if opts.adddsm
       ndsm = single(imread([opts.dataDir{1} 'ndsm/' ...
        opts.dataDir{3}{imToSample} '_normalized.jpg']))./255;
        im = cat(3,im,ndsm);
    end
    
    angle = (rand*360);

    im = gather(imrotate(gpuArray(im),angle));
    gt = gather(imrotate(gpuArray(gt),angle));
    [x,y] = getRandomPatches_1(im,w_pa,w_im,gt,w_gt,n,cep);
    x1 = cat(4,x1,x);
    y1 = cat(4,y1,y);
end

n = ceil(opts.batchSize*opts.NbatchVal/length(opts.valIms));

x2 = [];
y2 = [];
fprintf('v:')
for imToSample = opts.valIms
    fprintf('.')
    
    gt = single(rgb2label(imread([opts.dataDir{1} 'gts/' ...
        opts.dataDir{2}{imToSample} '.tif']),opts.legend));
    
    im = single(imread([opts.dataDir{1} 'tiles/' ...
        opts.dataDir{2}{imToSample} '.tif']))./255;
    
    if opts.adddsm
       ndsm = single(imread([opts.dataDir{1} 'ndsm/' ...
        opts.dataDir{3}{imToSample} '_normalized.jpg']))./255;
        im = cat(3,im,ndsm);
    end

    gt = single(rgb2label(imread([opts.dataDir{1} 'gts/' ...
        opts.dataDir{2}{imToSample} '.tif']),opts.legend));
    angle = (rand*360);

    im = gather(imrotate(gpuArray(im),angle));
    gt = gather(imrotate(gpuArray(gt),angle));
    [x,y] = getRandomPatches_1(im,w_pa,w_pa,gt,w_gt,n,cep);

    x2 = cat(4,x2,x);
    y2 = cat(4,y2,y);
end
fprintf('\n')

data = cat(4, x1, x2);
data(data>1) = 1;
data(data<0) = 0;
data(isnan(data)) = 0;

w_ = opts.predGtPatch;

imdb.images.data = data;
imdb.images.data_mean = opts.meanImage;
imdb.images.labels = double(cat(4, y1, y2));
imdb.images.set = [ones(1,size(y1,4)) 3*ones(1,size(y2,4))] ;
imdb.meta.sets = {'train', 'val', 'test'} ;
imdb.meta.classes = arrayfun(@(x)sprintf('%d',x),1:6,'uniformoutput',false) ;
imdb.meta.currentResolution = w_im;
imdb.gpus = opts.gpus;

end

function [x,y] = getRandomPatches_1(im,w_pa,w_im,gt,w_gt,num,cep)

if cep < 1
    class_to_exclude = [0 6];
else
    class_to_exclude = [0];
end
gt = single(gt);
ncl = numel(setdiff(unique(gt),class_to_exclude));

% possibility to sample following some prior over class frequency
if ncl == 5
    npc = ceil(num/ncl.*[1 1 1 1 1]);
    
elseif ncl == 6
    npc = ceil(num/ncl.*[1 1 1 1 1 1]);
    
end

cropPositions_im = cell(ncl,1);
cropPositions_gt = cell(ncl,1);

for class = setdiff(unique(gt),class_to_exclude)'
    mask = gather(ones(size(im,1),size(im,2)).*bwmorph(gpuArray(gt~=0),'erode',ceil(w_im/2)));
    mask = mask .* gt==class;
    
    [inz,jnz] = find(mask);
    
    ig = randperm(length(inz),npc(class))';
    
    cropPositions_im{class} = [inz(ig)-floor(w_im/2),jnz(ig)-floor(w_im/2),...
        inz(ig)+floor(w_im/2), jnz(ig)+floor(w_im/2)];
    
    cropPositions_gt{class} = [inz(ig)-floor(w_gt/2),jnz(ig)-floor(w_gt/2),...
        inz(ig)+floor(w_gt/2), jnz(ig)+floor(w_gt/2)];
    
end

cropPositions_im = cat(1,cropPositions_im{:});
cropPositions_gt = cat(1,cropPositions_gt{:});

y = gather(cropRectanglesMex_nd_nn(cat(3,gt,gt,gt), cropPositions_gt, [w_gt, w_gt]));
y = y(:,:,1,:);

x1 = gather(cropRectanglesMex_nd_cubic(im(:,:,1:3), cropPositions_im, [w_pa, w_pa]));
x2 = gather(cropRectanglesMex_nd_cubic(repmat(im(:,:,4),1,1,3), cropPositions_im, [w_pa, w_pa]));
x2 = x2(:,:,1,:);
x3 = [];

x = cat(3,x1,x2,x3);
x(x>1) = 1;
x(x<0) = 0;

end
