%%% USAGE:
% 1 - install matconvnet following instructions
% here [http://www.vlfeat.org/matconvnet/install/]
% 2 - Download (request) the Vaihingen or Potsdam datasets
% here [http://www2.isprs.org/commissions/comm3/wg4/tests.html]
% 3 - set up the folders from where to load datasets
% 4 - run
% PROTIP: read first comments below and possibly run some matconvnet
% examples (or check AT LEAST how matconvnet works!)
%
% Note 1 : if you'd like to retrieve probabilties rather than maps,
%     just add a variable output to DCNN_predict_denselabeling_cnn_blocks
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

addpath(genpath('../../../external_tools/matconvnet/matlab/'))
addpath(genpath('../../../external_tools/matconvnet/examples/'))
vl_setupnn();

addpath(genpath('./'))

% Colormap
opts.legend = [255 255 255;   % 1  Impervious surfaces
    0 0 255;       % 2  Buildings
    0 255 255;     % 3  Low vegetation
    0 255 0;       % 4  Tree
    255 255 0;     % 5  Car
    255 0 0;       % 6  Clutter / background
    ]./255;

%%

opts.dataset = 'Potsdam'; % or Potsdam

switch opts.dataset
    
    case 'Vaihingen'
        
        % e.g. from fine tuning:
        %         load('projects/cnn-deconv/todistribute/cnn_models/VA_fullPatchLabelingNet/net-epoch-2.mat')
        % the one from the paper:
        load('trained_networks/Vaihingen-denseprediction-distribution.mat')
        %
        net = dagnn.DagNN.loadobj(net) ;
        
        opts.adddsm    = true;
        opts.rgbMean   = [120.5403 81.7266 94.4010 80.5426]./255; %
        opts.saveDir   = ['maps_Vaihingen/']; % where to save maps and scores
        opts.imageDir  = ['../../../data/Vaihingen_dataset/images/tiles/']; % where are the image tiles of Vaihingen
        opts.ndsmDir   = ['../../../data/Vaihingen_dataset/images/ndsm/'];  % where are the ndsm tiles provided by M. Gerke
        opts.gtDir     = ['../../../data/Vaihingen_dataset/images/gts/'];   % where are the gt tiles
        
        valind = [11 15 28 30 34]; % indexes of Vaihingen images used for validation
        opts.vali_images = strsplit(sprintf('top_mosaic_09cm_area%i.tif ',valind));
        opts.vali_images = opts.vali_images(1:end-1);
        opts.vali_ndsm   = strsplit(sprintf('dsm_09cm_matching_area%i_normalized.jpg ',valind));
        opts.vali_ndsm   = opts.vali_ndsm(1:end-1);
        opts.ncut = 3;  % cut validation images to let them fit the GPU
        opts.gpus  = [1]; % let [] to remove use of GPU. [1 2] to use 2 GPUs etc
        
    case 'Potsdam'
        
        load('trained_networks/Potsdam-denseprediction-distribution.mat')
        net = dagnn.DagNN.loadobj(net) ;
        
        opts.adddsm    = true;
        opts.rgbMean   = [0.3375 0.3589 0.3350 0.3821 0.4918 ];
        
        % where to save, from where to load (asfor Vaihingen)
        opts.saveDir   = ['maps_Potsdam/'];
        opts.imageDir  = ['../../../data/Potsdam_dataset/images/4_Ortho_RGBIR/'];
        opts.ndsmDir   = ['../../../data/Potsdam_dataset/images/1_DSM_normalisation/'];
        opts.gtDir     = ['../../../data/Potsdam_dataset/images/5_Labels_for_participants/'];
        
        DCNN_setupPotsdam;
        
        opts.vali_images = opts_.valImages;
        opts.vali_ndsm   = opts_.valNdsm;
        clear opts_
        
        opts.ncut = 10; % cut validation images to let them fit the GPU
        opts.gpus  = [1]; % let [] to remove use of GPU. [1 2] to use 2 GPUs etc
        
end

likelihood = DCNN_predict_denselabeling_cnn_blocks(opts,net);

[SCORES] = DCNN_computeScoresFromMaps(opts.saveDir,opts.gtDir,opts.legend);

fprintf('full  GT scores :: OA %2.2f :: K %2.2f :: AA %2.2f :: mF1 %2.2f \n',...
    SCORES.fullGTallClassGlobal.OA,SCORES.fullGTallClassGlobal.Kappa,SCORES.fullGTallClassGlobal.AA,SCORES.fullGTallClassGlobal.avF1)
fprintf('nobkfuGT scores :: OA %2.2f :: K %2.2f :: AA %2.2f :: mF1 %2.2f \n',...
    SCORES.fullGTnoBackGlobal.OA,SCORES.fullGTnoBackGlobal.Kappa,SCORES.fullGTnoBackGlobal.AA,SCORES.fullGTnoBackGlobal.avF1)
fprintf('erodedGT scores :: OA %2.2f :: K %2.2f :: AA %2.2f :: mF1 %2.2f \n',...
    SCORES.noedgeGTallClassGlobal.OA,SCORES.noedgeGTallClassGlobal.Kappa,SCORES.noedgeGTallClassGlobal.AA,SCORES.noedgeGTallClassGlobal.avF1)
fprintf('nobkerGT scores :: OA %2.2f :: K %2.2f :: AA %2.2f :: mF1 %2.2f \n',...
    SCORES.noedgeGTnoBackGlobal.OA,SCORES.noedgeGTnoBackGlobal.Kappa,SCORES.noedgeGTnoBackGlobal.AA,SCORES.noedgeGTnoBackGlobal.avF1)
save([opts.saveDir '/SCORES.mat'],'SCORES')

% example plot: likelihood for class roads, for tile #3:
[~,labels] = max(likelihood(3).scores,[],3);
figure,
i1 = subplot(1,3,1);  subimage(likelihood(3).scores(:,:,1)); axis image, axis off
i3 = subplot(1,3,2);  subimage(labels==1); axis off
i2 = subplot(1,3,3);  subimage(labels,opts.legend); axis off
linkaxes([i1 i2 i3])



