function [SCORES] = DCNN_assess_multi_confmats(cmarray)

%%% compute figures of merit of multiple confusion matrices (sum(cm,3))
%%% output of assessment.m is compatible with cmarray(:,:,i).
%%% see assessment.m for further details. It has been modified to account
%%% for sparse class vectors and some further metrics added. Only
%%% constraints is that Lables are 1:[size(cmarray,1),size(cmarray,2)].

% Michele Volpi

% Compute confusion matrix
ConfusionMatrix = sum(cmarray,3);

PredUnique = sum(ConfusionMatrix,2)~=0;
TestUnique = find(sum(ConfusionMatrix,1)'~=0);

% Compute Overall Accuracy and Cohen's kappa statistic
n      = sum(ConfusionMatrix(:));                     % Total number of samples
PA     = sum(diag(ConfusionMatrix));
OA     = PA/n;

User  = diag(ConfusionMatrix)./sum(ConfusionMatrix,2).*100;
Prod  = (diag(ConfusionMatrix)'./sum(ConfusionMatrix,1))'.*100;

% Estimated Overall Cohen's Kappa (suboptimal implementation)
npj = sum(ConfusionMatrix,1);
nip = sum(ConfusionMatrix,2);
PE  = npj*nip;
if (n*PA-PE) == 0 && (n^2-PE) == 0
    % Solve indetermination
    warning('0 divided by 0')
    Kappa = 1;
else
    Kappa  = (n*PA-PE)/(n^2-PE);
end

% Store results:
SCORES.ConfusionMatrix = ConfusionMatrix;
SCORES.User            = User;
SCORES.avUser          = mean(User(PredUnique));
SCORES.Prod            = Prod;
SCORES.avProd          = mean(Prod(TestUnique));
SCORES.Kappa           = 100*Kappa;
SCORES.OA              = 100*OA;
SCORES.AA              = SCORES.avProd ;
SCORES.F1              = 2.*(SCORES.User.*SCORES.Prod)./(SCORES.User+SCORES.Prod);
SCORES.F1(isnan(SCORES.F1)) = 0;
SCORES.avF1            = mean(SCORES.F1(TestUnique));
end
