% As the name says
% give in varargin a list of GT ids to load if they don't match the file
% string pattern used to save maps
% 
% SCORES = 
% 
%             fullGTallClass: Per-image scores, all classes, all boundaries, {1 X #valIms}
%               fullGTnoBack: Per-image scores, without background, all boundaries, {1 X #valIms}
%           noedgeGTallClass: Per-image scores, all classes, 3px eroded boundary, {1 X #valIms}
%             noedgeGTnoBack: Per-image scores, without background, 3px eroded boundary, {1 X #valIms}
%       fullGTallClassGlobal: same as fullGTallClass, but for pooled confusion matrix
%         fullGTnoBackGlobal: same as fullGTnoBack, but for pooled confusion matrix
%     noedgeGTallClassGlobal: same as oedgeGTallClass, but for pooled confusion matrix
%       noedgeGTnoBackGlobal: same as noedgeGTnoBack, but for pooled confusion matrix
% 
%               Michele Volpi

function [SCORES] = DCNN_computeScoresFromMaps(folderMaps,folderGTs,legend)


if ~(exist(folderMaps,'dir') == 7)
    error('folderMaps = %s is not a folder',folderMaps)
end

if ~(exist(folderGTs,'dir') == 7)
    error('folderMaps = %s is not a folder',folderGTs)
end

fileList = dir(folderMaps);
fileList = fileList(~[fileList.isdir]);
fileList = {fileList.name};
strs = strfind(fileList,'MAP');
for s = 1:length(strs)
    if ~isempty(strs{s})
        tempList{s} = fileList{s};
    end
end
fileList = tempList;

CONFMAT_fu = zeros(6,6,length(fileList));
CONFMAT_er = zeros(6,6,length(fileList));

for imid   = 1:length(fileList)

    pred   = imread([folderMaps fileList{imid}]);
    if size(pred,3) == 1
        pred   = rgb2label(rgb2label(pred+1,legend),legend);
    else
        pred = rgb2label(pred,legend);
    end
    
    if ~isempty(strfind(lower(folderGTs),'potsdam'))
        gts    = rgb2label(imread([folderGTs fileList{imid}(5:end-16) '_label.tif' ]),legend);
    elseif ~isempty(strfind(lower(folderGTs),'vaihingen'))
        gts    = rgb2label(imread([folderGTs strrep(strrep(fileList{imid},'MAP-',''),'_class','')]),legend);
    else
        error('cannot recognize data undefined')
    end
    
    gt_c = gts(:);
    pr_c = pred(:);
    
    PredUnique = unique(pred);
    TestUnique = unique(gts);
    
    CONFMAT_fu(:,:,imid) = accumarray([pr_c,gt_c],1,[6 6]);
    
    SCORES.fullGTallClass{imid} = DCNN_assess_multi_confmats(CONFMAT_fu(:,:,imid));
    SCORES.fullGTnoBack{imid}   = DCNN_assess_multi_confmats(CONFMAT_fu(1:5,1:5,imid));

    borders = ~imdilate(edge(gts),strel('disk',3));
    gt_e = gt_c.*borders(:);
    gt_c(gt_e == 0) = [];
    pr_c(gt_e == 0) = [];
    
    for i = 1:length(PredUnique)
        for j = 1:length(TestUnique)
            CONFMAT_er(PredUnique(i),TestUnique(j),imid) = ...
                sum(pr_c==PredUnique(i) & gt_c==TestUnique(j));
        end
    end

    SCORES.noedgeGTallClass{imid} = DCNN_assess_multi_confmats(CONFMAT_er(:,:,imid));
    SCORES.noedgeGTnoBack{imid}   = DCNN_assess_multi_confmats(CONFMAT_er(1:5,1:5,imid));
    
end

SCORES.fullGTallClassGlobal   = DCNN_assess_multi_confmats(CONFMAT_fu);
SCORES.fullGTnoBackGlobal     = DCNN_assess_multi_confmats(CONFMAT_fu(1:5,1:5,:));
SCORES.noedgeGTallClassGlobal = DCNN_assess_multi_confmats(CONFMAT_er);
SCORES.noedgeGTnoBackGlobal   = DCNN_assess_multi_confmats(CONFMAT_er(1:5,1:5,:));

%      4916494      116782      281147       69918      193795        6387
%       341505     5489298      351633       81047       12987        3134
%       414035      108359     3669413      719028        8920        2626
%       204939       55566      962578     5121641        6189         222
%         5045         145          61           8       57178        1080
%          494           0           0           0           0        2140
    
    