function  probabilties = DCNN_predict_denselabeling_cnn_blocks(opts,net)

%%% predicts labeling map using the network specified in net,
%%% with options specified in opts.
%%% Results are saved in opts.saveDir
%%% for more details see MatConvNet manual and code!!
% 1 - install matconvnet following instructions
% here [http://www.vlfeat.org/matconvnet/install/]
% 2 - Download (request) the Vaihingen or Potsdam datasets
% here [http://www2.isprs.org/commissions/comm3/wg4/tests.html]
% 3 - set up the folders from where to load datasets
% 4 - run
% PROTIP: read first comments below and possibly run some matconvnet
% examples (or check AT LEAST how matconvnet works!)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% the following is if you fine-tuned the net. Need to replace last layers
if ~strcmp(net.getLayer(length(net.layers)).name,'softmax')
    for name = {'objective', 'accuracy'};%,'top2err'}
        net.removeLayer(name) ;
    end
    net.addLayer('softmax', dagnn.SoftMax(),'prediction','softmax')
end

predVar = net.getVarIndex('softmax') ;

% setup stuff
inputVar = 'input' ;
net.mode = 'test' ;

if ~isempty(opts.gpus)
    net.move('gpu')
else
    net.move('cpu')
end

pad = [65 65];

mkdir(opts.saveDir)

for id = 1:length(opts.vali_images)
    
    TESTIM  = char(opts.vali_images(id));
    TESTDSM = char(opts.vali_ndsm(id));
    
    IM   = single(imread([char(strcat(opts.imageDir,TESTIM))]))./255;
    nDSM = single(imread([char(strcat(opts.ndsmDir,TESTDSM))]))./255;
    
    IMs_g = cat(3,IM,nDSM);
    
    fprintf(': %s :',char(TESTIM))
    
    Prob_w = zeros(size(IM,1),size(IM,2),6,'single');
    
    
    %     cut images into pieces, I've reached my last resort. to fit GPU memory
    cutY = [1 [1:opts.ncut].*floor(size(IM,2)/opts.ncut)];
    cutX = [1 [1:opts.ncut].*floor(size(IM,1)/opts.ncut)];
    
    
    for kY = 1:(length(cutY)-1)
        fprintf('.')
        for kX = 1:(length(cutX)-1)
            IMs = IMs_g(cutX(kX):cutX(kX+1),cutY(kY):cutY(kY+1),:);
            sizeIM = size(IMs);
            
            IMs = padarray(IMs,[pad(1) pad(2)],'post','symmetric');
            
            switch opts.dataset
                case 'Vaihingen'
                    meanImage    = single(cat(3,repmat(opts.rgbMean(1),size(IMs,1),size(IMs,2)), ...
                        repmat(opts.rgbMean(2),size(IMs,1),size(IMs,2)),...
                        repmat(opts.rgbMean(3),size(IMs,1),size(IMs,2)),...
                        repmat(opts.rgbMean(4),size(IMs,1),size(IMs,2))));%...
                case 'Potsdam'
                    meanImage    = single(cat(3,repmat(opts.rgbMean(1),size(IMs,1),size(IMs,2)), ...
                        repmat(opts.rgbMean(2),size(IMs,1),size(IMs,2)),...
                        repmat(opts.rgbMean(3),size(IMs,1),size(IMs,2)),...
                        repmat(opts.rgbMean(4),size(IMs,1),size(IMs,2)),...
                        repmat(opts.rgbMean(5),size(IMs,1),size(IMs,2))));%...
            end
            
            % center data on training set
            IMs = IMs - meanImage;
            clear meanImage
            
            if ~isempty(opts.gpus)
                IMs = gpuArray(single(IMs));
            else
                IMs = single(IMs);
            end
            
            % do forward pass in the new
            net.eval({inputVar, IMs}) ;
            
            Prob = gather(net.vars(predVar).value) ;
            
            clear res
            % stitch together tiles
            Prob_w(cutX(kX):cutX(kX+1),cutY(kY):cutY(kY+1),:) = Prob(1:sizeIM(1),1:sizeIM(2),:);
        end
    end
    
    [~,MAP] = max(Prob_w,[],3);
    
    if ~isempty(nargout)
        probabilties(id).scores = Prob_w;
    end
    
    figure; imshow(MAP,opts.legend);
    
    imwrite(MAP,opts.legend,[char(strcat(opts.saveDir,'MAP-',TESTIM(1:end-4),'_class.tif'))],'compression','none')
    clear IMs_g
    fprintf('\n')
end

